//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

// These functions are used in properly displaying textures over the wrapped
// board surface. Only display4() and rdisplay4() should be called.
//
// x/y min/max determine where the screen edges map to the board.
// x and y are the board coordinates of the texture.
// r is the radius of the texture to be rendered.
// texture is the texture to be rendered.
//
// The functions adjust the board coordinates to the appropriate screen
// coordinates.  When there is ambiguity at the edges, the texture will be
// be displayed on both (or, at the corners, all four) sides.
//
// The 'r' versions of these functions will additionally rotate each displayed
// texture counterclockwise around its center by theta radians.
//
// Programmer's Note:
// I have tested a number of methods of rendering. Of the methods that
// rendered properly at the edges, this one seems to be the quickest, at least
// on my machine. The culling gives more advantage than the many, many
// conditionals might counteract.

#ifndef GG_GLSHARED_H
#define GG_GLSHARED_H

#include"shared.h"
#include<GL/gl.h>

inline void display1(double x, double y, double r, const CenteredTexture &texture)
{
  glPushMatrix();
  glTranslated(x,y,0);
  glScaled(r,r,1);
  texture.display();
  glPopMatrix();
}

inline void display2(double ymin, double ymax, double x, double y, double r, const CenteredTexture &texture)
{
  if(y < ymin + r)
  {
    if(y >= ymin - r)
      display1(x,y,r,texture);
    y += V_RESOLUTION;
  }
  else if(y >= ymax - r)
  {
    if(y < ymax + r)
      display1(x,y,r,texture);
    y -= V_RESOLUTION;
  }
  display1(x,y,r,texture);
}

inline void display4(double xmin, double xmax, double ymin, double ymax, double x, double y, double r, const CenteredTexture &texture)
{
  if(x < xmin + r)
  {
   if(x >= xmin - r)
      display2(ymin,ymax,x,y,r,texture);
    x += H_RESOLUTION;
  }
  else if(x >= xmax - r)
  {
    if(x < xmax + r)
      display2(ymin,ymax,x,y,r,texture);
    x -= H_RESOLUTION;
  }
  display2(ymin,ymax,x,y,r,texture);
}

inline void rdisplay1(double x, double y, double r, double theta, const CenteredTexture &texture)
{
  glPushMatrix();
  glTranslated(x,y,0);
  glScaled(r,r,1);
  glRotated(theta,0,0,1);
  texture.display();
  glPopMatrix();
}

inline void rdisplay2(double ymin, double ymax, double x, double y, double r, double theta, const CenteredTexture &texture)
{
  if(y < ymin + r)
  {
    if(y >= ymin - r)
      rdisplay1(x,y,r,theta,texture);
    y += V_RESOLUTION;
  }
  else if(y >= ymax - r)
  {
    if(y < ymax + r)
      rdisplay1(x,y,r,theta,texture);
    y -= V_RESOLUTION;
  }
  rdisplay1(x,y,r,theta,texture);
}

inline void rdisplay4(double xmin, double xmax, double ymin, double ymax, double x, double y, double r, double theta, const CenteredTexture &texture)
{
  if(x < xmin + r)
  {
   if(x >= xmin - r)
      rdisplay2(ymin,ymax,x,y,r,theta,texture);
    x += H_RESOLUTION;
  }
  else if(x >= xmax - r)
  {
    if(x < xmax + r)
      rdisplay2(ymin,ymax,x,y,r,theta,texture);
    x -= H_RESOLUTION;
  }
  rdisplay2(ymin,ymax,x,y,r,theta,texture);
}

#endif
