//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

#include"shared.h"
#include"mixer.h"
#include"highscore.h"

#include"texture.h"
#include"scrolltexture.h"
#include"fonttexture.h"
#include"centeredtexture.h"

#include"worm.h"
#include"apple.h"
#include"powerup.h"
#include"visual.h"

#include<SDL/SDL.h>
#include<GL/gl.h>

#include<time.h> // time()
#include<stdlib.h> // srand()

#include<iostream>
#include<sstream>
#include<list>
#include<string>

enum {
  gameStateTitleWarm,
  gameStateTitleCool,
  gameStatePause,
  gameStatePlay,
  gameStateHighScore,
  gameStateOver
} gameState = gameStateTitleWarm;

SDL_Surface *screen;
bool fullscreen;

Texture title;
ScrollTexture backd;
ScrollTexture backn;
ScrollTexture front;
FontTexture font_texture;        // global in main.h
CenteredTexture head_texture;    // global in main.h
CenteredTexture segment_texture; // global in main.h
CenteredTexture apple_texture;   // global in main.h
CenteredTexture powerup_texture; // global in main.h

unsigned long int last_frame; // marks time
bool left_key  = false;       // indicates whether the specified key is held down
bool right_key = false;       //  "
double direction_delta = 0.0; // total direction change from user input

Worm worm;
std::list<Collider*> colliders;
std::list<Visual*> visuals;

HighScoreList highscores("annelid.dat");
std::string placement; // used when getting the user's name
std::string name;      // the user's name

#define HELP_PAGES 3
int  helppage = 0;
bool helphide = false;

void unload_textures()
{
  title.unload();
  backd.unload();
  backn.unload();
  front.unload();
  font_texture.unload();
  head_texture.unload();
  segment_texture.unload();
  apple_texture.unload();
  powerup_texture.unload();
}

void clear_game_lists()
{
  for(std::list<Collider*>::iterator cp = colliders.begin(); cp != colliders.end(); cp = colliders.erase(cp))
    delete (*cp);
  for(std::list<Visual*>::iterator vp = visuals.begin(); vp != visuals.end(); vp = visuals.erase(vp))
    delete (*vp);
}

void setup_new_game()
{
  // clear out the objects from the previous game, if there are any
  clear_game_lists();

  // generate a new worm and apple
  worm = Worm();
  colliders.push_back(new Apple());
}

void state_change()
{
  switch(gameState)
  {
    case gameStateTitleWarm: // user wants to hurry up and play
    {
      gameState = gameStateTitleCool;

      // set up the title fade to fade out from the current point in the fade in
      unsigned long int now = SDL_GetTicks();
      if(now - last_frame < 1000)
        last_frame = 2*now - 1000 - last_frame; // set the start of the cool down in the past
      else
        last_frame = now;
      break;
    }

    case gameStateTitleCool: // they can wait for this!
      break;

    case gameStateHighScore: // user is done entering a name or doesn't want to do so
      highscores.add(worm.score(),name.c_str());
      gameState = gameStateOver;
      break;

    case gameStatePlay: // user wants to pause
#ifndef DEBUG_NO_GRABBING
      SDL_ShowCursor(1);
      SDL_WM_GrabInput(SDL_GRAB_OFF);
#endif
      gameState = gameStatePause;
      break;

    case gameStateOver: // user wants to start a new game
      setup_new_game();
      // fall through

    case gameStatePause: // user wants to unpause
#ifndef DEBUG_NO_GRABBING
      SDL_ShowCursor(0);
      SDL_WM_GrabInput(SDL_GRAB_ON);
#endif
      gameState = gameStatePlay;
      last_frame = SDL_GetTicks();
      break;
  }
}

bool handle_input()
{
  SDL_Event event;

  while(SDL_PollEvent(&event))
  {
    switch(event.type)
    {
      case SDL_KEYDOWN:
        switch(event.key.keysym.sym)
        {
          case SDLK_ESCAPE:
            return false;

          case SDLK_LEFT:  left_key  = true; break;
          case SDLK_RIGHT: right_key = true; break;

          case SDLK_PAGEUP:
            if(gameState == gameStatePause)
              if(--helppage < 0)
                helppage = HELP_PAGES - 1;
            break;

          case SDLK_PAGEDOWN:
            if(gameState == gameStatePause)
              if(++helppage >= HELP_PAGES)
                helppage = 0;
            break;

          case SDLK_TAB:
            if(gameState == gameStatePause)
              helphide = !helphide;
            break;

          // Thanks to pasp (http://clay.ll.pl/) for pointing this functionality out
          case SDLK_F2:
            SDL_WM_ToggleFullScreen(screen);
            fullscreen = !fullscreen; // this may be useful later, but is unused now
            break;

#ifdef DEBUG_EXTRA_KEYS
          case SDLK_HOME: worm.grow(1);  break;
          case SDLK_END:  worm.grow(-1); break;

          case SDLK_0:
          case SDLK_1:
          case SDLK_2:
          case SDLK_3:
          case SDLK_4:
          case SDLK_5:
          case SDLK_6:
          case SDLK_7:
          case SDLK_8:
          case SDLK_9:
            if(gameState != gameStateHighScore)
            {
              worm.affect(EffectType(event.key.keysym.sym-SDLK_0));
              break;
            }
            // else fall through
#endif

          case SDLK_F1:    // in case the user expects help
          case SDLK_SPACE:
            if(gameState != gameStateHighScore)
            {
              state_change();
              break;
            }
            // else fall through

          default:
            if(gameState == gameStateHighScore) // entering a name
            {
              if(event.key.keysym.sym >= 'a' && event.key.keysym.sym <= 'z' && name.size() < MAX_NAMESIZE)
              {
                char letter = event.key.keysym.sym;
                if(event.key.keysym.mod & (KMOD_CAPS | KMOD_SHIFT) &&
                   !(event.key.keysym.mod & KMOD_CAPS && event.key.keysym.mod & KMOD_SHIFT))
                   letter += 'A' - 'a'; // capitalize
                name += letter;
              }
              else if(event.key.keysym.sym == ' ' && name.size() < MAX_NAMESIZE)
                name += ' ';
              else if(event.key.keysym.sym == SDLK_BACKSPACE && name.size())
                name.resize(name.size()-1);
              else if(event.key.keysym.sym == SDLK_RETURN)
                state_change();
            }
            break;
        }
        break;

      case SDL_KEYUP:
        switch(event.key.keysym.sym)
        {
          case SDLK_LEFT:  left_key  = false; break;
          case SDLK_RIGHT: right_key = false; break;

          default:
            break;
        }
        break;

      case SDL_MOUSEBUTTONDOWN:
        state_change();
        break;

      case SDL_MOUSEMOTION:
        if(gameState == gameStatePlay)
          direction_delta -= event.motion.xrel * MOUSE_SENSITIVITY * PI / 180.0;
        break;

      case SDL_QUIT:
        return false;
    }
  }

  return true;
}

int main(int argc, char *argv[])
{
  // check the command line for options
  // these are technically undocumented
  if(argc > 1)
  {
    for(int i = 1; i < argc; i++)
    {
      std::string option(argv[i]);
      if(option == "version"   ||
         option == "-v"          )
      {
        std::cout << "Analog Annelid" << std::endl
                  << "Version " << VERSION_STRING << std::endl
                  << "Uses SDL under the GNU LGPL." << std:: endl
                  << "For more information please see readme.txt or" << std::endl
                  << "Visit http://www.garthgillespie.com/annelid/" << std::endl
                  << " Copyright 2004 Garth Gillespie, All Rights Reserved." << std::endl
                  << std::endl;
        return 0;
      }
      else if(option == "fullscreen" ||
              option == "-fs"          )
        fullscreen = true;
    }
  }

  // initialize sdl
  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0)
  {
    std::cerr << "Error: Worm: " << SDL_GetError() << std::endl;
    return 1;
  }

  // initialize sdl screen resource
  SDL_WM_SetCaption("Analog Annelid","Worm");
  if(SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1) ||
     (screen = SDL_SetVideoMode (H_RESOLUTION,V_RESOLUTION,16,SDL_DOUBLEBUF | SDL_OPENGL | (fullscreen?SDL_FULLSCREEN:0))) == NULL)
  {
    std::cerr << "Error: Worm: " << SDL_GetError() << std::endl;
    SDL_Quit();
    return 2;
  }

  // load textures
  if(!title.load("title.tga") ||
     !backd.load("backd.tga") ||
     !backn.load("backn.tga") ||
     !front.load("front.tga") ||
     !font_texture.load("font") ||
     !head_texture.load("head.tga") ||
     !segment_texture.load("segment.tga") ||
     !apple_texture.load("apple.tga") ||
     !powerup_texture.load("powerup.tga"))
  {
    std::cerr << "Error: Worm: Could not open graphics." << std::endl;
    unload_textures();
    SDL_Quit();
    return 3;
  }

  // initialize the mixer system
  if(Mixer_init() < 0)
    std::cerr << "Error: Worm: Mixer: " << SDL_GetError() << std::endl;

  // register sounds
  // this is not strictly necessary, but it should speed the program's
  // response the first time a sound is played
  Mixer_reg("score.wav");
  Mixer_reg("apple.wav");
  Mixer_reg("speed.wav");
  Mixer_reg("slow.wav");
  Mixer_reg("grow.wav");
  Mixer_reg("shrink.wav");
  Mixer_reg("vortex.wav");

  // set up gl environment
  glClearColor(0,0,0,0);
  glMatrixMode(GL_MODELVIEW);
  glEnable(GL_BLEND);
  glBlendFunc (GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);

  // set up the worm, apple, etc.
  srand(time(0));
  setup_new_game();
  last_frame = SDL_GetTicks();

  // while input is ok, update the game and then the graphics
  while(handle_input())
  {
    // if not paused, update the worm
    if(gameState == gameStatePlay      ||
       gameState == gameStateHighScore ||
       gameState == gameStateOver        )
    {
      // Complete frame updates until caught up with the current time.
      // This can cause jittery graphics with poor graphic acceleration.
      // This can also lock up very slow computers if it takes longer than
      // TICKS_per_FRAME to do a single update.
      // However, it will ensure constant motion on a suitable computer.
      for(unsigned long int now = SDL_GetTicks();
          last_frame +  TICKS_per_FRAME < now;
          last_frame += TICKS_per_FRAME)
      {
        // update direction with keyboard input
        if(left_key)  direction_delta += STEERING_ANGLE;
        if(right_key) direction_delta -= STEERING_ANGLE;

        // move the worm one frame
        if(!worm.update(direction_delta,visuals) &&
           gameState == gameStatePlay              ) // the worm has died
        {
          // alter game state
#ifndef DEBUG_NO_GRABBING
          SDL_ShowCursor(1);
          SDL_WM_GrabInput(SDL_GRAB_OFF);
#endif

          // determine if a high score was achieved
          unsigned int place = highscores.place(worm.score());
          if(place)
          {
            std::ostringstream sout;

            sout << "A score of " << worm.score() << " makes " << place;
            if     (place%10 == 1 && place != 11) sout << "st";
            else if(place%10 == 2 && place != 12) sout << "nd";
            else if(place%10 == 3 && place != 13) sout << "rd";
            else                                  sout << "th";
            sout << " place on the high score list!\nPlease enter your name.\n";
            placement = sout.str();

            gameState = gameStateHighScore;
          }
          else
            gameState = gameStateOver;
        }

        // potentially spawn new powerups
        PowerUp *pu = new PowerUp();
        if(pu->valid())
          colliders.push_back(pu);
        else
          delete pu;

        // update the colliders, and see if they have been hit
        for(std::list<Collider*>::iterator cp = colliders.begin(); cp != colliders.end();)
        {
          if(!(*cp)->update()) // collider has ceased to exist (a powerup or vortex)
          {
            delete (*cp);
            cp = colliders.erase(cp);
          }

          else if(worm.collides((*cp)->x(),(*cp)->y(),(*cp)->r()))
          {
            (*cp)->collided(worm,colliders,visuals);

            delete (*cp);
            cp = colliders.erase(cp);
          }

          else
           cp++;
        }

        // update the visual effects
        for(std::list<Visual*>::iterator vp = visuals.begin(); vp != visuals.end();)
        {
          if(!(*vp)->update()) // visual has finished
          {
            delete (*vp);
            vp = visuals.erase(vp);
          }
          else
            vp++;
        }
      }
    }

    // clear the screen
    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();
    glTranslated(-1,-1,0);
    glScaled(2,2,0);

    if(gameState == gameStateTitleWarm ||
       gameState == gameStateTitleCool   )
    {
      unsigned long int now = SDL_GetTicks();
      double alpha;

      if(gameState == gameStateTitleWarm)
        alpha = (now - last_frame < 1000)
              ? (now - last_frame) / 1000.0
              : 1.0;
      else // gameState == gameStateTitleCool
      {
        if(now - last_frame < 1000)
          alpha = 1.0 - (now - last_frame) / 1000.0;
        else
        {
          alpha = 0;
          last_frame = now;
          gameState = gameStatePause;
        }
      }

      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
      glColor4d(1,1,1,alpha);
      title.display();
      glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);
    }
    else
    {
      // display scrolling backgrounds (stars and nebulae)
      backd.display(worm.scrollx()/H_RESOLUTION/2.0,worm.scrolly()/V_RESOLUTION/2.0);
      backn.display(worm.scrollx()/H_RESOLUTION/1.5,worm.scrolly()/V_RESOLUTION/1.5);

      // shift the screen so that the worm's head is center screen
      glScaled(1.0/H_RESOLUTION,1.0/V_RESOLUTION,1); // point per pixel
      glTranslated(H_RESOLUTION/2.0 - worm.x(),V_RESOLUTION/2.0 - worm.y(),0);

      // determine the screen boundaries for quick display of wrapping elements
      double xmin = worm.x() - H_RESOLUTION/2.0;
      double xmax = worm.x() + H_RESOLUTION/2.0;
      double ymin = worm.y() - V_RESOLUTION/2.0;
      double ymax = worm.y() + V_RESOLUTION/2.0;

      // display the worm, colliders, and visual effects
      // note: they will be layered in that order bottom to top, in the order they were created
      worm.display(xmin,xmax,ymin,ymax);
      for(std::list<Collider*>::iterator cp = colliders.begin(); cp != colliders.end(); cp++)
        (*cp)->display(xmin,xmax,ymin,ymax);
      for(std::list<Visual*>::iterator vp = visuals.begin(); vp != visuals.end(); vp++)
        (*vp)->display(xmin,xmax,ymin,ymax);

      // display scrolling foreground (cloud cover)
      glLoadIdentity();
      glTranslated(-1,-1,0);
      glScaled(3,3,0); // *1.5
      front.display(worm.scrollx()/H_RESOLUTION,worm.scrolly()/V_RESOLUTION);

      // setup display for text
      glLoadIdentity();
      glTranslated(-1,1,0);
      glScalef(32.0/H_RESOLUTION,32.0/V_RESOLUTION,1);
      glTranslatef(0,-1,0);

      if(gameState == gameStateOver)
      {
        // display high scores
        highscores.display();
      }
      else if(gameState == gameStateHighScore)
      {
        // display a message asking for input and relay that input
        font_texture.display(placement.c_str());
        font_texture.display(name.c_str());
      }
      else if(gameState == gameStatePause && !helphide)
      {
        // dull the game graphics while paused
        glDisable(GL_TEXTURE_2D);
        glColor4f(0.5,0.5,0.5,0.5);
        glBegin(GL_QUADS);
          glVertex3f( 0  ,-31.0,0);
          glVertex3f(32.0,-31.0,0);
          glVertex3f(32.0,  1.0,0);
          glVertex3f( 0  ,  1.0,0);
        glEnd();
        glEnable(GL_TEXTURE_2D);

        // display the appropriate help message
        switch(helppage)
        {
          case 0:
            // starting help message, controls
            glTranslatef(2.0,1.0-(V_RESOLUTION/16.0-14.0)/2.0,0); // note: 14 is the lines of text in the message below
            font_texture.display(
              "Controls:\n"
              "\n"
              "mouse left/right or keyboard [Left]/[Right]:\n"
              "steers the worm CCW/CW\n"
              "\n"
              "mouse button or [Space]:\n"
              "starts, pauses, or unpauses the game\n"
              "\n"
              "[Esc]:\n"
              "quits the game\n"
              "\n"
              "\n"
              "Press [PgUp] or [PgDn] for more help,\n"
              "Press [Tab] to hide/show this message.\n"
            );
            break;

          case 1:
          {
            // help message on colliders
            static StaticVisual     apple_visual(0.5,0.5,0.5,&apple_texture);
            static SpinFasterVisual faster_visual(0.5,0.5,0.5,&segment_texture);
            static SpinSlowerVisual slower_visual(0.5,0.5,0.5,&segment_texture);
            static GrowerVisual     grower_visual(0.5,0.5,0.5,&segment_texture);
            static ShrinkerVisual   shrinker_visual(0.5,0.5,0.5,&segment_texture);
            static RotatorVisual    vortex_visual(0.5,0.5,0.5,&powerup_texture);

            apple_visual.update();
            faster_visual.update();
            slower_visual.update();
            grower_visual.update();
            shrinker_visual.update();
            vortex_visual.update();

            glTranslatef(2.0,1.0-(V_RESOLUTION/16.0-13.0)/2.0,0); // note: 13 is the lines of text in the message below
            font_texture.display("Objects:\n\n");
            apple_visual.display(-2,2,-2,2);
            font_texture.display("      Apple: collect these for points and length.\n\n");
            faster_visual.display(-2,2,-2,2);
            font_texture.display("      Speeding Segment: speeds the worm up.\n\n");
            slower_visual.display(-2,2,-2,2);
            font_texture.display("      Slowing Segment: slows the worm down.\n\n");
            grower_visual.display(-2,2,-2,2);
            font_texture.display("      Growing Segment: makes the worm wider.\n\n");
            shrinker_visual.display(-2,2,-2,2);
            font_texture.display("      Shrinking Segment: makes the worm thinner.\n\n");
            vortex_visual.display(-2,2,-2,2);
            font_texture.display("      Swirling Vortex: has a random, wacky effect!\n\n");
            break;
          }

          case 2:
            // miscellaneous information
            glTranslatef(2.0,1.0-(V_RESOLUTION/16.0-14.0)/2.0,0); // note: 14 is the lines of text in the message below
            font_texture.display(
              "Analog Annelid is the creation of Garth Gillespie.\n"
              "\n"
              "It may be redistributed in its original .zip file\n"
              "without modificaton and as long as no profit is\n"
              "made from its distribution.\n"
              "\n"
              "The sounds in Analog Annelid are public domain.\n"
              "\n"
              "Analog Annelid uses SDL under the GNU LGPL.\n" // this is required for a "work that uses the Library"
              "\n"
              "For more information see readme.txt or...\n"
              "Please visit http://www.garthgillespie.com/annelid/\n"
              "\n"
              " Copyright 2004 Garth Gillespie, All Rights Reserved.\n"
            );
        }
      }
      else // gameState == gameStatePlay || (gameState == gameStatePause && helphide)
      {
        // display current score
        std::ostringstream sout;
        sout << "Score: " << worm.score();
        font_texture.display(sout.str().c_str());
      }
    }

    // flush graphics to screen
    glFlush();
    SDL_GL_SwapBuffers();
  }

  // exiting, clean up
  highscores.write();
  highscores.close();
  clear_game_lists();
  unload_textures();
  Mixer_down();
  SDL_Quit();

  return 0;
}
