//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

#include"mixer.h"

#include<list>
#include<string>
#include<iostream>

#include<SDL/SDL.h>
#include<string.h> // memset()
#include<math.h> // atan()

#define PI 3.1415926535897932384626433832795
#define ATAN_SCALE (PI/2.0/32767.0)

namespace gg_mixer
{
  // status of the mixer subsystem
  bool valid = false;

  // this class represents sound data
  class Sound
  {
    public:
      Sound(const char *fname):name(fname)
      {
        SDL_AudioSpec wav_spec;
        unsigned char *wav_stream;
        if(!SDL_LoadWAV(fname,&wav_spec,&wav_stream,&length))
        {
          std::cerr << "Error: Mixer: Could not load " << fname << ": " << SDL_GetError() << std::endl;
          data = 0;
          return;
        }

        // FIXME: freq interpolation/extrapolation
        switch(wav_spec.format)
        {
          case AUDIO_U8: // 8bit unsigned
          {
            // set up data members to hold the wav data
            // length /= sizeof(Uint8);
            length /= wav_spec.channels;
            data = new double[length];

            // copy and convert the data, scaled to -PI/2 < x < PI/2 for use with arctangent
            Uint8 *wav_data = (Uint8*)(wav_stream);
            if(wav_spec.channels == 1)
              for(unsigned int i = 0; i < length; i++)
                data[i] = (double(wav_data[i])-128.0)*(PI/2.0/127.0);
            else
              for(unsigned int i = 0; i < length; i++)
                data[i] = (double(wav_data[2*i])+double(wav_data[2*i+1])-256.0)*(PI/2.0/127.0);

            break;
          }

          case AUDIO_S8: // 8bit signed
          {
            // set up data members to hold the wav data
            // length /= sizeof(Sint8);
            length /= wav_spec.channels;
            data = new double[length];

            // copy and convert the data, scaled to -PI/2 < x < PI/2 for use with arctangent
            Sint8 *wav_data = (Sint8*)(wav_stream);
            if(wav_spec.channels == 1)
              for(unsigned int i = 0; i < length; i++)
                data[i] = double(wav_data[i])*(PI/2.0/127.0);
            else
              for(unsigned int i = 0; i < length; i++)
                data[i] = (double(wav_data[2*i])+double(wav_data[2*i+1]))*(PI/2.0/127.0);

            break;
          }

          case AUDIO_U16LSB: // 16bit unsigned
          case AUDIO_U16MSB:
          {
            // set up data members to hold the wav data
            length /= sizeof(Uint16);
            length /= wav_spec.channels;
            data = new double[length];

            // if the byte order doesn't match the system, swap the bytes in the stream
            if(wav_spec.format != AUDIO_U16SYS)
              for(unsigned int i = 0; i < length; i += 2)
              {
                int tmp = wav_stream[i];
                wav_stream[i] = wav_stream[i+1];
                wav_stream[i+1] = tmp;
              }

            // copy and convert the data, scaled to -PI/2 < x < PI/2 for use with arctangent
            Uint16 *wav_data = (Uint16*)(wav_stream);
            if(wav_spec.channels == 1)
              for(unsigned int i = 0; i < length; i++)
                data[i] = (double(wav_data[i])-32768.0)*(PI/2.0/32767.0);
            else
              for(unsigned int i = 0; i < length; i++)
                data[i] = (double(wav_data[2*i])+double(wav_data[2*i+1])-65536.0)*(PI/2.0/32767.0);

            break;
          }

          case AUDIO_S16LSB: // 16bit signed
          case AUDIO_S16MSB:
          {
            // set up data members to hold the wav data
            length /= sizeof(Sint16);
            length /= wav_spec.channels;
            data = new double[length];

            // if the byte order doesn't match the system, swap the bytes in the stream
            if(wav_spec.format != AUDIO_S16SYS)
              for(unsigned int i = 0; i < length; i += 2)
              {
                int tmp = wav_stream[i];
                wav_stream[i] = wav_stream[i+1];
                wav_stream[i+1] = tmp;
              }

            // copy and convert the data, scaled to -PI/2 < x < PI/2 for use with arctangent
            Sint16 *wav_data = (Sint16*)(wav_stream);
            if(wav_spec.channels == 1)
              for(unsigned int i = 0; i < length; i++)
                data[i] = double(wav_data[i])*(PI/2.0/32767.0);
            else
              for(unsigned int i = 0; i < length; i++)
                data[i] = double(double(wav_data[2*i])+double(wav_data[2*i+1]))*(PI/2.0/32767.0);

            break;
          }

          default:
            std::cerr << "Error: Mixer: " << fname << " Bad sound format." << std::endl;
            data = 0;
            break;
        }

        SDL_FreeWAV(wav_stream);
      }

      bool         valid()  const { return data; }
      unsigned int duration() const { return length; }
      double*      stream() const { return data; }
      void free() { delete [] data; }

      bool operator==(const char *fname) const { return name == fname; }

    private:
      std::string name;
      unsigned int length;
      double *data;
  };

  // this class represents a playing sound
  class Voice
  {
    public:
      Voice(Sound *s):sound(s),position(0) {}

      Voice& operator++(int) { position++; return *this; }

      bool playing() const { return position < sound->duration(); }
      double sample() const { return sound->stream()[position]; } // FIXME: error control?

    private:
      Sound *sound;
      unsigned int position;
  };

  SDL_AudioSpec hardware; // FIXME: is it necessary to store this value?
  std::list<Voice> voices;
  std::list<Sound> sounds;

  // SDL calls this function periodically to update the sound stream
  void callback(void *userdata, Uint8 *stream, int len)
  {
    // if all voices are done, audio is not needed
    if(!voices.size())
    {
      SDL_PauseAudio(0);
      return;
    }

    // FIXME: variable hardware
    len /= sizeof(Sint16);

    // allocate an array of doubles for mixing
    double *dstream = new double[len];
    memset(dstream,0,len*sizeof(double));

    // mix the voices into a single stream
    for(std::list<Voice>::iterator v = voices.begin(); v != voices.end();)
    {
      // add the voice to the stream
      for(int i = 0;
          i < len && (*v).playing();
          i++, (*v)++)
        dstream[i] += (*v).sample(); // -PI/2 < x < PI/2

      // if voice is finished, drop it
      if(!(*v).playing())
        v = voices.erase(v);
      else
        v++;
    }

    // convert the mixing stream to the hardware stream
    Sint16 *hardware_stream = (Sint16*)(stream); // 16-bit mono
    for(int i = 0; i < len; i++)
      hardware_stream[i] = Sint16(atan(dstream[i])/ATAN_SCALE); // clamp to 16-bit

    delete [] dstream;
  }
}

using namespace gg_mixer; // FIXME: is this necessary?

int Mixer_init()
{
  // ensure that the sound is not up
  Mixer_down();

  // set up a request for a sound format
  SDL_AudioSpec desired;
  desired.freq = 44100;
  desired.format = AUDIO_S16SYS;
  desired.channels = 1;
  desired.samples = 8192;
  desired.callback = callback;
  desired.userdata = 0;

  // alert calling function that there was an error
  if(SDL_OpenAudio(&desired,&hardware) < 0 ||
     hardware.format != AUDIO_S16SYS         )
    return -1;

  // alert the mixer system and the calling function
  valid = true;
  return 0;
}

void Mixer_down()
{
  // if sound is not up anyway, there is nothing to do
  if(!valid)
    return;

  // free memory allocated by Sounds and erase the list
  for(std::list<Sound>::iterator s = sounds.begin(); s != sounds.end(); s = sounds.erase(s))
    (*s).free();

  // erase the Voices
  voices.clear();

  // shutdown sound system
  SDL_CloseAudio();

  // alert the mixer system
  valid = false;
}

void Mixer_reg(const char *fname)
{
  // if sound is not up, do nothing
  if(!valid)
    return;

  // if the sound already exists, do nothing
  for(std::list<Sound>::iterator s = sounds.begin(); s != sounds.end(); s++)
    if(*s == fname)
      return;

  // load the sound and add it to the list
  sounds.push_back(Sound(fname));
}

void Mixer_add(const char *fname)
{
  // if sound is not up, do nothing
  if(!valid)
    return;

  std::list<Sound>::iterator s;

  // find the appropriate sound
  for(s = sounds.begin(); s != sounds.end(); s++)
    if(*s == fname)
    {
      // if the sound is invalid, do nothing
      if((*s).valid())
        break;
      else
        return;
    }

  // if the appropriate sound was not found, load it
  if(s == sounds.end())
  {
    s = sounds.insert(s,Sound(fname));
    // if the sound is invalid, do nothing
    if(!(*s).valid())
      return;
  }

  // add it to the mix
  SDL_LockAudio();
  voices.push_back(Voice(&*s));
  SDL_UnlockAudio();

  // ensure that the mixer callback function is called
  SDL_PauseAudio(0);
}
