//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

#ifndef GG_UTILITY_H
#define GG_UTILITY_H

#include"shared.h"

// align_quadrant() returns the equivalent gamespace value of offset o2 that
// is in the same quadrant as o1, based on a gamespace range of res (H_RESOLUTION
// or V_RESOLUTION).  This ensures the shortest actual distance (given wrapping)
// between o1 and the return value.
inline double align_quadrant(double o1, double o2, double res)
{
  if(o1 - o2 < -res/2.0) return o2 - res;
  if(o1 - o2 >  res/2.0) return o2 + res;
  return o2;
}


// collision() determines whether two circular areas overlap in gamespace.
inline bool collision(double x1, double y1, double r1,
                      double x2, double y2, double r2)
{
  // find the shortest horizontal and vertical distances between the centers
  double dx = x1 - align_quadrant(x1,x2,H_RESOLUTION);
  double dy = y1 - align_quadrant(y1,y2,V_RESOLUTION);

  // find the greatest distance at which a collision occurs
  double d = r1 + COLLISION_FACTOR*r2;

  // calculate if the actual distance is close enough to collide
  return dx*dx + dy*dy <= d*d;
}


// segments_intersect() determines if the line segment defined by p1 and p2
// intersects with the line segment defined by p3 and p4 .
// This algorithm was found on the Internet, and is well-known.
//
// direction_of_point() determines the direction of a point from the line
// segment defined by p1 and p2 .
//
// point_on_segment() determines if a point p which is collinear with p1 and
// p2 lies on the segment defined by p1 and p2 .

inline double direction_of_point(double px,  double py,
                                 double p1x, double p1y,
                                 double p2x, double p2y)
{
  // return : (p - p1) x (p2 - p1)
  // pa - pb = {pax-pbx,pay-pby}
  // pa x pb = pax*pby - pbx*pay
  return (px-p1x)*(p2y-p1y) - (p2x-p1x)*(py-p1y);
}

inline bool point_on_segment(double px,  double py,
                             double p1x, double p1y,
                             double p2x, double p2y)
{
  // return: px is between p1x and p2x, and py is between p1y and p2y
  return (p1x < p2x ? (p1x <= px && px <= p2x) : (p2x <= px && px <= p1x) ) &&
         (p1y < p2y ? (p1y <= py && py <= p2y) : (p2y <= py && py <= p1y) )    ;
}

inline bool segments_intersect(double p1x, double p1y,
                               double p2x, double p2y,
                               double p3x, double p3y,
                               double p4x, double p4y)
{
  double d1 = direction_of_point(p1x,p1y,p3x,p3y,p4x,p4y);
  double d2 = direction_of_point(p2x,p2y,p3x,p3y,p4x,p4y);
  double d3 = direction_of_point(p3x,p3y,p1x,p1y,p2x,p2y);
  double d4 = direction_of_point(p4x,p4y,p1x,p1y,p2x,p2y);

  // if the signs of d1 and d2 and the signs of d3 and d4 are different, then
  // the two line segments fundamentally intersect.
  if(d1*d2 < 0 && d3*d4 < 0)
    return true;

  // if any of the directions are 0, then that point is collinear with the
  // other line segment.  if the point is actually on that line segment, then
  // the two segments intersect at an endpoint.
  if(d1 == 0 && point_on_segment(p1x,p1y,p3x,p3y,p4x,p4y)) return true;
  if(d2 == 0 && point_on_segment(p2x,p2y,p3x,p3y,p4x,p4y)) return true;
  if(d3 == 0 && point_on_segment(p3x,p3y,p1x,p1y,p2x,p2y)) return true;
  if(d4 == 0 && point_on_segment(p4x,p4y,p1x,p1y,p2x,p2y)) return true;

  // otherwise, the segments do not intersect
  return false;
}

#endif
