//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

#include"visual.h"

///////////////////////////////////////

#include"glshared.h" // includes <GL/gl.h>

void StaticVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  display4(xmin,xmax,ymin,ymax,x,y,r*mult,*texture); // display with appropriate location, wrapping, and size
}

void RotatorVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  // display with appropriate rotation, location, wrapping, and size (one rotation every two seconds)
  rdisplay4(xmin,xmax,ymin,ymax,x,y,r*mult,frame*180.0/FRAMES_per_SECOND,*texture);
}

///////////////////////////////////////

//#include"glshared.h" // includes <GL/gl.h>
#include<math.h> // sin(),cos()

void SpinFasterVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  double theta = frame*2.0*PI*SPIN_FAST_REVS/FRAMES_per_SECOND; // determine the angle of revolution based on the age
  double rm = r*mult*0.75; // scale the image down a bit to better fit the actual size
  display4(xmin,xmax,ymin,ymax,x+rm*cos(theta),y+rm*sin(theta),rm,*texture); // display with appropriate location, wrapping, and size
}

void SpinSlowerVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  double theta = frame*2.0*PI*SPIN_SLOW_REVS/FRAMES_per_SECOND; // determine the angle of revolution based on the age
  double rm = r*mult*0.75; // scale the image down a bit to better fit the actual size
  display4(xmin,xmax,ymin,ymax,x+rm*cos(theta),y+rm*sin(theta),rm,*texture); // display with appropriate location, wrapping, and size
}

///////////////////////////////////////

//#include"glshared.h" // includes <GL/gl.h>

void GrowerVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  // display with appropriate location, wrapping, and size (based on age)
  double rm = GROW_SHRINK_MIN + (GROW_SHRINK_MAX-GROW_SHRINK_MIN) * double(frame < (GROW_SHRINK_FRAMES/2) ? frame : (GROW_SHRINK_FRAMES/2))/(GROW_SHRINK_FRAMES/2);
  display4(xmin,xmax,ymin,ymax,x,y,r*mult*rm,*texture);
}

void ShrinkerVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  // display with appropriate location, wrapping, and size (based on age)
  double rm = GROW_SHRINK_MAX - (GROW_SHRINK_MAX-GROW_SHRINK_MIN) * double(frame < (GROW_SHRINK_FRAMES/2) ? frame : (GROW_SHRINK_FRAMES/2))/(GROW_SHRINK_FRAMES/2);
  display4(xmin,xmax,ymin,ymax,x,y,r*mult*rm,*texture);
}

///////////////////////////////////////

//#include"glshared.h" // includes <GL/gl.h>

void GrowAndFadeVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);           // enable sprite color modulation, for alpha
  glColor4d(1,1,1,1.0 - double(frame) / GROW_FADE_DURATION);           // appropriate transparency
  visual->multiplier((1.0 + double(frame) / GROW_FADE_DURATION)*mult); // appropriate size
  visual->display(xmin,xmax,ymin,ymax);                                // display the child visual
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_REPLACE);            // disable sprite color modulation
}

///////////////////////////////////////

//#include"shared.h"

//#include<math.h>   // sin(),cos(),sqrt()
#include<sstream>

FlyingNumberVisual::FlyingNumberVisual(double theta, unsigned long int nin, FontTexture *fin)
:x(H_RESOLUTION/2.0),y(V_RESOLUTION/2.0-8.0),font(fin)
{
  // convert the number to a string
  std::ostringstream sout;
  sout << nin;
  number = sout.str();

  // adjust x so that the number is centered (numbers in the selected font are all 8 pixels wide)
  x -= double(number.length())*4.0;

  // determine the x and y velocities from the direction parameter
  dx = FLYING_NUMBER_VELOCITY*cos(theta);
  dy = FLYING_NUMBER_VELOCITY*sin(theta);

  // some math:
  // newtonian physics (position from acceleration): x = x0 + t * v0 + 0.5 * a * t^2
  // plug in vertical movement from center screen: 0.0 = -(V_RESOLUTION/2.0) + (dy) * t + 0.5 * t^2
  //
  // quadratic equation: x = (-b (+/-) sqrt(b*b - 4*a*c)) / (2*a)
  // t = -dy (+/-) sqrt(dy*dy + V_RESOLUTION)
  // note dy*dy+V_RESOLUTION is always positive, so there's always a real answer
  // also, addition is always preferred, since negative time is irrelevant
  //
  // ax = (56.0 - H_RESOLUTION/2.0 - dx * t)/(0.5 * t*t) => [reworked]
  // ax = (112.0 - H_RESOLUTION - 2.0*dx*t)/(t*t)
  // together, all of this will make the number end its journey at the top
  // of the screen and 56 pixels right of the left edge

  double t = -dy + sqrt(dy*dy + V_RESOLUTION);
  ax = (112.0 - H_RESOLUTION - 2.0*dx*t)/(t*t);
}

bool FlyingNumberVisual::update()
{
  // accelerate movement and adjust location
  x += dx += ax;
  y += dy += 1.0; // 'ay' always equals 1.0

  // if the location is off of the screen, the visual is finished
  if(y > V_RESOLUTION)
    return false;
  return true;
}

void FlyingNumberVisual::display(double xmin, double xmax, double ymin, double ymax) const
{
  glPushMatrix();

  // remove the adjustment to center the worm's head on the screen and adjust to the location of the visual
  // (i.e. lock the visual to the screen, not the play surface)
  glTranslated(xmin+x,ymin+y,0);

  glScaled(16.0*mult,16.0*mult,0); // characters are 16 pixels tall, does not preserve centering if 'mult' is used
  font->display(number.c_str());
  glPopMatrix();
}
