//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

// The Visual class is used for various visual effects.
//
// update() is called to let the effect change over time, and to alert the
// program when it has finished.
//
// display() displays the effect, its parameters define the viewable portion
// of game space.
//
// Static displays a texture at a given location and size.
//
// Rotator displays a texture of a given location and size that rotates around
// its center.
//
// SpinFast displays a texture that orbits a given point at a given size,
// counterclockwise.  The speed of orbiting is defined by SPIN_FAST_REVS in
// revolutions per second.
//
// SpinSlow displays a texture that orbits a given point at a given size,
// clockwise.  The speed of orbiting is defined by SPIN_SLOW_REVS in
// revolutions per second.
//
// Grower displays a texture at given point that grows from small to large,
// pausing for a moment at its largest size.
//
// Shrinker displays a texture at given point that shrinks from large to
// small, pausing for a moment at its smallest size.
//
// The Grower and Shrinker Visuals pulse in a cycle GROW_SHRINK_FRAMES long.
// Half of the time is spent growing or shrinking, the other half pausing.
// GROW_SHRINK_MAX and GROW_SHRINK_MIN determine the large and small sizes
// respectively, as a multiplier.
//
// GrowAndFade displays another, given Visual that grows to double size while
// fading to complete transparency in time specified by GROW_FADE_DURATION .
//
// FlyingNumber produces a textual number that arcs toward the score region of
// the screen.  Its initial direction is passed to the constructor; its
// initial velocity is specified by FLYING_NUMBER_VELOCITY .

#ifndef GG_VISUAL_H
#define GG_VISUAL_H

class Visual
{
  public:
    Visual():mult(1) {}

    virtual ~Visual() {}
    virtual bool update() = 0; // returns false when finished, not all visuals finish
    virtual void display(double xmin, double xmax, double ymin, double ymax) const = 0;

    void multiplier(double min) { mult = min; }

  protected:
    double mult; // this is used for nesting Visuals that need to alter the size of children
};

///////////////////////////////////////

#include"centeredtexture.h"

class StaticVisual:public Visual
{
  public:
    // pass original board coordinates, size, and texture
    StaticVisual(double xin, double yin, double rin, CenteredTexture *tin)
    :x(xin),y(yin),r(rin),texture(tin) {}

    virtual bool update() { return true; } // lasts forever
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // board coordinates
    double y;
    double r; // starting radius
    CenteredTexture *texture;
};

class RotatorVisual:public Visual
{
  public:
    // pass original board coordinates, size, and texture
    RotatorVisual(double xin, double yin, double rin, CenteredTexture *tin)
    :x(xin),y(yin),r(rin),frame(0),texture(tin) {}

    virtual bool update() { frame++; return true; } // lasts forever
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // board coordinates
    double y;
    double r; // starting radius
    unsigned int frame;
    CenteredTexture *texture;
};

///////////////////////////////////////

//#include"centeredtexture.h"
#include"shared.h"

#define SPIN_FAST_REVS 4.0
#define SPIN_SLOW_REVS 1.0

class SpinFasterVisual:public Visual
{
  public:
    // pass original board coordinates, size, and texture
    SpinFasterVisual(double xin, double yin, double rin, CenteredTexture *tin)
    :x(xin),y(yin),r(rin),frame(1),texture(tin) {}

    virtual bool update() { ++frame; return true; } // lasts forever
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // board coordinates
    double y;
    double r; // starting radius
    unsigned int frame;
    CenteredTexture *texture;
};

class SpinSlowerVisual:public Visual
{
  public:
    // pass original board coordinates, size, and texture
    SpinSlowerVisual(double xin, double yin, double rin, CenteredTexture *tin)
    :x(xin),y(yin),r(rin),frame(FRAMES_per_SECOND),texture(tin) {}

    virtual bool update() { --frame; return true; } // lasts forever
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // board coordinates
    double y;
    double r; // starting radius
    unsigned int frame;
    CenteredTexture *texture;
};

///////////////////////////////////////

//#include"centeredtexture.h"
//#include"shared.h"

#define GROW_SHRINK_FRAMES (2*FRAMES_per_SECOND)
#define GROW_SHRINK_MAX    1.2
#define GROW_SHRINK_MIN    0.8

class GrowerVisual:public Visual
{
  public:
    // pass original board coordinates, size, and texture
    GrowerVisual(double xin, double yin, double rin, CenteredTexture *tin)
    :x(xin),y(yin),r(rin),frame(0),texture(tin) {}

    virtual bool update() { if(++frame > GROW_SHRINK_FRAMES) frame = 0; return true; } // lasts forever
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // board coordinates
    double y;
    double r; // starting radius
    unsigned int frame;
    CenteredTexture *texture;
};

class ShrinkerVisual:public Visual
{
  public:
    // pass original board coordinates, size, and texture
    ShrinkerVisual(double xin, double yin, double rin, CenteredTexture *tin)
    :x(xin),y(yin),r(rin),frame(0),texture(tin) {}

    virtual bool update() { if(++frame > GROW_SHRINK_FRAMES) frame = 0; return true; } // lasts forever
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // board coordinates
    double y;
    double r; // starting radius
    unsigned int frame;
    CenteredTexture *texture;
};

///////////////////////////////////////

//#include"shared.h"

// lasts one half second
#define GROW_FADE_DURATION (FRAMES_per_SECOND/2)

class GrowAndFadeVisual:public Visual
{
  public:
    // pass the child visual, which will be destroyed upon completion
    GrowAndFadeVisual(Visual *vin)
    :frame(0),visual(vin) {}

    ~GrowAndFadeVisual() { delete visual; }

    virtual bool update() { return (visual->update() && ++frame > GROW_FADE_DURATION) ? false : true; } // last for a predetermined duration
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    unsigned int frame;
    Visual *visual;
};

///////////////////////////////////////

#include"fonttexture.h"

#include<string>

#define FLYING_NUMBER_VELOCITY 8.0

class FlyingNumberVisual:public Visual
{
  public:
    // pass initial direction, number, and font texture
    // initial location is assumed to be center screen
    FlyingNumberVisual(double theta, unsigned long int nin, FontTexture *fin);

    virtual bool update(); // lasts until the number flies off of the screen vertically
    virtual void display(double xmin, double xmax, double ymin, double ymax) const;

  private:
    double x; // screen coordinates
    double y;
    double dx; // direction deltas
    double dy;
    double ax; // acceleration values
//    double ay; // always == 1.0
    std::string number;
    FontTexture *font;
};

#endif
