//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

#include"worm.h"
#include"main.h"
#include"glshared.h" // includes "shared.h", <GL/gl.h>

#include<stdlib.h> // rand(),RAND_MAX
#include<math.h> // sin(),cos(),atan2(),sqrt()


bool Worm::update(double &dd, std::list<Visual*> &visuals)
{
  if(dead)
  {
    // annihilate the body with a gradual visual effect
    if(body.size())
    {
      Segment s = body.front();
      visuals.push_back(new GrowAndFadeVisual(new StaticVisual(s.x,s.y,s.r,&segment_texture)));
      body.pop_front();
    }

    // adjust head's position
    head.x += speed*cos(direction);
    head.y += speed*sin(direction);

    // adjust scroll position
    scroll_x += speed*cos(direction);
    scroll_y += speed*sin(direction);

    return false;
  }

  // apply all timed effects
  for(std::list<TimedEffect>::iterator te = timedeffects.begin(); te != timedeffects.end();)
  {
    if(++(*te).frame < EFFECT_DURATION)
    {
      switch((*te).type)
      {
        case effectJitter:
          dd += JITTER_RANGE * RANDOM_FLOAT() - JITTER_RANGE/2.0;
          break;

        case effectSwap:
          dd = -dd;
          break;

        case effectDeform:
        {
          // get a random value in a multiplicatively balanced range around 1
          double r = 2.0 * DEFORM_FACTOR * RANDOM_FLOAT() + 1.0 - DEFORM_FACTOR;
          if(r > 1.0)
            r = 1.0 / (2.0 - r);

          head.r *= r;
//          speed *= r;
          limit();
          break;
        }

        default:
          break;
      }
      te++;
    }
    else
      te = timedeffects.erase(te);
  }

  // alter direction and clear dd
  direction += dd;
  dd = 0.0;

  // if it is time to move to a new segment, push it on
  // if it is time to loose a segment, pop one off
  if(++frame >= FRAMES_per_SEGMENT)
  {
    body.push_front(head);
    if(pending_segments > 0) // growing
      pending_segments--;
    else
    {
      body.pop_back();
      if(pending_segments < 0 && body.size()) // shrinking, don't allow popping when there is nothing to pop
      {
        body.pop_back();
        pending_segments++;
      }
    }
    frame = 0;
  }

  // adjust head's position
  head.x += speed*cos(direction);
  head.y += speed*sin(direction);

  // wrap x if necessary
  if(head.x >= H_RESOLUTION) head.x -= H_RESOLUTION;
  else if(head.x < 0)        head.x += H_RESOLUTION;

  // wrap y if necessary
  if(head.y >= V_RESOLUTION) head.y -= V_RESOLUTION;
  else if(head.y < 0)        head.y += V_RESOLUTION;

  // adjust scrolling offset
  scroll_x += speed*cos(direction);
  scroll_y += speed*sin(direction);

  if(body.size() > 1)
  {
    std::list<Segment>::iterator s = body.begin();
    // point f is used with head to form a line segment representing the head
    // align all points used in these calculations to one (the head's) quadrant
    double fx = align_quadrant(head.x,(*s).x,H_RESOLUTION);
    double fy = align_quadrant(head.y,(*s).y,V_RESOLUTION);

    s++; // skip the first body line segment, as it always intesects the head line segment at its endpoint
    // point t is used with point s below to form a line segment representing the body
    double tx = align_quadrant(head.x,(*s).x,H_RESOLUTION);
    double ty = align_quadrant(head.y,(*s).y,V_RESOLUTION);

    // skip the initial overlapping segments, making sure the worm does not cross itself
    for(s++; s != body.end() && collision((*s).x,(*s).y,(*s).r,head.x,head.y,head.r); s++)
    {
      // point s is used with point t to form a line segment representing the body
      double sx = align_quadrant(head.x,(*s).x,H_RESOLUTION);
      double sy = align_quadrant(head.y,(*s).y,V_RESOLUTION);

      if(segments_intersect(head.x,head.y,fx,fy,tx,ty,sx,sy)) // if the head and body line segments cross
      {
        dead = true;
        return false;
      }

      tx = sx;
      ty = sy;
    }

    // check if the worm has collided with any of its remaining body segments
    for(; s != body.end(); s++)
      if(collision((*s).x,(*s).y,(*s).r,head.x,head.y,head.r))
      {
        dead = true;
        return false;
      }
  }

  return true; // always alive here
}

void Worm::limit()
{
  // clamp size
  if(head.r > MAX_SEGMENT_RADIUS)      head.r = MAX_SEGMENT_RADIUS;
  else if(head.r < MIN_SEGMENT_RADIUS) head.r = MIN_SEGMENT_RADIUS;

  // clamp speed, based on size
  double max_speed = 2.0*head.r/FRAMES_per_SEGMENT; // never move so fast that there are gaps
  if(speed > max_speed)      speed = max_speed;
  else if(speed < MIN_SPEED) speed = MIN_SPEED;
}

void Worm::affect(EffectType effect)
{
  switch(effect)
  {
    case effectFast:
      speed *= EFFECT_RANGE * RANDOM_FLOAT() + EFFECT_MIN;
      limit();
      break;

    case effectSlow:
      speed /= EFFECT_RANGE * RANDOM_FLOAT() + EFFECT_MIN;
      limit();
      break;

    case effectThick:
      head.r *= EFFECT_RANGE * RANDOM_FLOAT() + EFFECT_MIN;
      limit();
      break;

    case effectThin:
      head.r /= EFFECT_RANGE * RANDOM_FLOAT() + EFFECT_MIN;
      limit();
      break;

    case effectReverse:
      if(body.size()) // if there is a body, swap the head with the last segment
      {
        body.push_front(head); // make the head a segment
        head = body.back();    // make the last segment the head
        body.pop_back();

        // get the vector between what will be the new first segment and the head
        double dx = head.x - align_quadrant(head.x,body.back().x,H_RESOLUTION);
        double dy = head.y - align_quadrant(head.y,body.back().y,V_RESOLUTION);

        // determine the direction and speed of the vector, and thus the worm
        direction = atan2(dy,dx);
        if(body.size() != 1)
          frame = FRAMES_per_SEGMENT;
        speed = sqrt(dx*dx + dy*dy)/frame;

        limit();        // just in case something is amiss
        body.reverse(); // reverse the body
      }
      else // if there is no body, just reverse direction
        direction += PI;
      break;

    case effectJitter:
    case effectSwap:
    case effectDeform:
      timedeffects.push_back(TimedEffect(effect));
      break;

    default:
      break;
  }
}

void Worm::display(double xmin, double xmax, double ymin, double ymax) // const
{
  // display the body (or what's left of it, if dying)
  for(std::list<Segment>::reverse_iterator s = body.rbegin(); s != body.rend(); s++)
    display4(xmin,xmax,ymin,ymax,(*s).x,(*s).y,(*s).r,segment_texture);

  // if alive, display the head, only one is needed since it is always center screen
  if(!dead)
  {
    glPushMatrix();
    glTranslated(head.x,head.y,0);
    glRotated(180.0 / PI * direction,0,0,1);
    glScaled(head.r,head.r,1);
    head_texture.display();
    glPopMatrix();
  }
}
