//  Copyright 2004 Garth Gillespie, All Rights Reserved.
// This C++ code is part of Analog Annelid, the creation of Garth Gillespie,
// And may only be used for educational, non-profit purposes.
// For more information, please visit http://www.garthgillespie.com/annelid/ .

// The Worm class represents the the worm as it moves through gamespace,
// running into various things.  Worm also keeps track of the scroll offsets
// for the foreground and backgrounds, and the player's score.
//
// update() is called to allow the worm to move.  It returns false when the
// worm is dead.  The parameter direction_delta is the amount of steering
// since the last update, update sets it to 0. The parameter visuals allows
// update() to add visual effects to the scene, specifically the worm death
// effects.
//
// display() displays the worm, its parameters define the viewable portion
// of game space.
//
// x() and y() return the position of the worm's head.
// dir() returns the worm's direction.
// scrollx() and scrolly() return the current scroll offsets.
//
// collides() determines if an object at position x,y with radius r overlaps
// the worm's head.
//
// grow() will alert the worm to gain or lose segments as the game progresses.
// affect() will alter the worm with a particular effect, such as a powerup.
//
// score() will return the player's current score, or alter it by i.

#ifndef GG_WORM_H
#define GG_WORM_H

#include"shared.h"
#include"visual.h"
#include"utility.h"

#include<list>

class Worm
{
  public:
    Worm():head(0,0,INITIAL_RADIUS),direction(PI/2.0),speed(STARTING_SPEED),
           scroll_x(0),scroll_y(0),frame(FRAMES_per_SEGMENT),
           pending_segments(STARTING_SEGMENTS),dead(false),points(0) {}

    bool update(double &direction_delta, std::list<Visual*> &visuals);

    void display(double xmin, double xmax, double ymin, double ymax); // const;

    double x()   const { return head.x; }
    double y()   const { return head.y; }
    double dir() const { return direction; }

    double scrollx() const { return scroll_x; }
    double scrolly() const { return scroll_y; }

    bool collides(double x, double y, double r) const { return !dead && collision(x,y,r,head.x,head.y,head.r); }

    void grow(int segments) { pending_segments += segments; }
    void affect(EffectType effect);

    unsigned long int score() const     { return points; }
    void              score(long int i) { points += i; }

  private:
    void limit(); // constrains speed and head size to reasonable limits

    // TimedEffect keeps track of how long certain effects have gone on
    struct TimedEffect
    {
      TimedEffect(EffectType effect):type(effect),frame(0) {}

      EffectType type;
      unsigned int frame;
    };

    // a Segment is a single portion of the worm
    struct Segment
    {
      Segment(double xin, double yin, double rin):x(xin),y(yin),r(rin) {}

      double x;
      double y;
      double r;
    };

    Segment head;
    std::list<Segment> body;

    double direction;
    double speed;

    double scroll_x;
    double scroll_y;

    int frame;
    std::list<TimedEffect> timedeffects;

    int pending_segments;
    bool dead;

    unsigned long int points;
};

#endif
